/************************************************************************

    SCANDAL v1.00 - Prime Ministerial Scandal Simulator

    Copyright (C) 2004 John Elliott <jce@seasip.demon.co.uk>

    This software is provided "as is," without warranty of any kind, express
    or implied.  In no event shall the author be held liable for any direct, 
    indirect, incidental, special or consequential damages arising out of the 
    use of or inability to use this software.

    Permission is granted to anyone to use this software for any purpose,
    including commercial applications, and to alter it and redistribute it
    freely, subject to the following restrictions:

    1. Redistributions of source code must retain the above copyright notice,
       definition, disclaimer, and this list of conditions.

    2. Redistributions in binary form must reproduce the above copyright
       notice, definition, disclaimer, and this list of conditions in
       documentation and/or other materials provided with the distribution.

   3. Altered versions--including, but not limited to, ports to new operating
      systems, existing ports with new graphical interfaces, and dynamic,
      shared, or static library versions--must be plainly marked as such
      and must not be misrepresented as being the original source. 

*************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

static int screen_w = 40;

/* If compiling with Pacific C under DOS, go for the full works with 
 * flashy colours and so forth */
#ifdef __PACIFIC__

#include <conio.h>
#include <dos.h>

/* Check for "Sinclair" anywhere in the ROM BIOS */
int is_sinclair()
{
	int n;
	static char rom[16384];
	char far *ptr;
      
	ptr = MK_FP(0xFC00, 0);
	for (n = 0; n < 16384; n++) rom[n] = ptr[n];

	for (n = 0; n < 16376; n++)
	{
		if (!memcmp(rom + n, "Sinclair", 8)) return 1;
	}
	return 0;
}

static char far *framebuf;
static int mode;

/* Clear screen */
void cls()
{
	union REGS rg;
	rg.x.ax = mode & 0xFF;
	int86(0x10, &rg, &rg);
}

/* Detect mono or colour, & behave appropriately */
void initscr()
{
	union REGS rg;

	rg.x.ax = 0x0F00;
	int86(0x10, &rg, &rg);
	if (rg.h.al == 7) 
	{
		mode = 7;
		screen_w = 80;
		framebuf = MK_FP(0xB000, 0);
	}
	else
	{
		mode = 0;
		screen_w = 40;
		framebuf = MK_FP(0xB800, 0);
	}
	cls();
}


/* Draw a character (by poking it into the framebuffer) */
static char monoattr[] = { 7, 15, 9, 1, 0x7f, 0x70 };

void drawchar(int row, int col, char c, int bold)
{
	int offset = (row * screen_w + col) * 2;
	int ink, paper;
	static int lb, lc;

	framebuf[offset    ] = c;

	if (mode == 7)	framebuf[offset + 1] = monoattr[bold];
	else	
	{
		if (!bold) framebuf[offset + 1] = 7;
		else if (bold == lb)
		{
			framebuf[offset + 1] = lc;
		}
		else
		{
			paper = rand() % 8;
			do
			{
				ink = (rand() % 16);
			}
			while (((ink & 7) )  == paper || 
			       ((ink & 7)+1) == paper ||
			       ((ink & 7)-1) == paper ||
			       ((ink & 7)+2) == paper ||
			       ((ink & 7)-2) == paper);
			framebuf[offset+1] = lc = paper*16 + ink;
		}
	}
	lb = bold;
}
#else

/* UNIX interface (for testing) */
#include <termios.h>
#include <unistd.h>

int is_sinclair()
{
	return 0;
}

void cls()
{
	int n;
	for (n = 0; n < 24; n++) putchar('\n');
}

int getch()
{
	struct termios ts, ots;
	char c;
	int i;
	int filen = fileno(stdin);

	tcgetattr(filen, &ts);
	tcgetattr(filen, &ots);
	cfmakeraw(&ts);
	tcsetattr(filen, TCSANOW, &ts);
	fflush(stdout);
	fflush(stdin);
	fflush(stderr);
	do
	{
		i = read(filen, &c, 1);
	} while (i < 0);
	tcsetattr(filen, TCSANOW, &ots);
	fflush(stdin);
	return c;
}

static int curcol = 0;

void initscr()
{
	screen_w = 80;
	cls();
}

void drawchar(int row, int col, char c, int bold)
{
	if (col < curcol) 
	{ 
		putchar('\n'); 
		curcol = 0; 
	}
	while (curcol < col) 
	{ 
		putchar(' '); 
		++curcol; 
	}
	if (bold)
	{
		putchar('_');
		putchar(8);
	}
	putchar(c);
	++curcol;
	fflush(stdout);
}

#endif
/* End of machine-dependent bits */

/* How many characters will a string occupy when printed? */
int rlen(char *s)
{
	int n = 0;

	for (n = 0; *s; s++)
	{
		if (strchr("@~#*+-", *s)) continue;
		++n;
	}
	return n;
}

/* Print a string, wrapping words at the end of a line 
 * 
 * Uses a simple highlighting system, in which the characters
 * @ ~ # * +  turn on highlight
 * -          turns off highlight
 *
 * On a colour display, any highlighted section displays in colours chosen
 * pretty much at random. On a mono display, the highlights correspond to 
 * particular attributes.
 */
void printat(int row, int col, const char *fmt, ...)
{
	va_list ap;
	char buf[4096], *s, *t;
	int bold = 0, wlen;

	va_start(ap, fmt);
	vsprintf(buf, fmt, ap);
	va_end(ap);

	for (s = buf; s[0]; s++)
	{
		if (*s == '@') { bold = 5; continue; }
		if (*s == '~') { bold = 4; continue; }
		if (*s == '#') { bold = 3; continue; }
		if (*s == '*') { bold = 2; continue; }
		if (*s == '+') { bold = 1; continue; }
		if (*s == '-') { bold = 0; continue; }
		if (*s == '\n') { col = 0; ++row; continue; }
		t = strchr(s, ' ');
		if (t) 
		{
			char c = *t;
			*t = 0;
			wlen = rlen(s);
			*t = c;
		}
		else   wlen = rlen(s);

		if (wlen > (screen_w - 1- col))
		{
			col = 0;
			++row;
		}

		drawchar(row, col++, s[0], bold);
	}
}

/* Centre a string */
void centre(int row, const char *s, int bold)
{
	switch(bold)
	{
		case 5: printat(row, (screen_w - strlen(s)) / 2, "@%s-", s);
			break;
		case 4: printat(row, (screen_w - strlen(s)) / 2, "~%s-", s);
			break;
		case 3: printat(row, (screen_w - strlen(s)) / 2, "#%s-", s);
			break;
		case 2: printat(row, (screen_w - strlen(s)) / 2, "*%s-", s);
			break;
		case 1: printat(row, (screen_w - strlen(s)) / 2, "+%s-", s);
			break;
		default:printat(row, (screen_w - strlen(s)) / 2, "%s", s);
			break;
	}
}

/* Your poll rating */
static int rating = 50;

/* Introduction screen */
void intro()
{
	centre(0, "Scandal", 1);	
	centre(2, "A ground-breaking political satire", 0);
	centre(3, "(and not at all a low-budget ripoff of", 0);
	centre(4, "other people's ideas or current events,", 0);
	centre(5, "oh, dear me, no)", 0);
	printat(7, 0, "  Congratulations! You've become the #Prime Minister!- "
		      "Fame, celebrity, tea with the +Queen- and maybe even a "
		      "little *real power- await!\n"
		      "  +But- there's a slight problem. Your #ministers- "
		      "are a bunch of hopeless layabouts only in it for the "
		      "+free drinks- and @loose women-. Can you keep them out "
		      "of trouble?");

	centre(24, "Press ENTER to start", 4);
	getch();
}

/* List of ministers. The first character is:
 * _ Can be male or female 
 * m Must be male
 * f Must be female
 *
 * Note that the game writes to these strings; so if building with gcc, you
 * have to use -fwritable-strings */

static char *ministers[] =
{
	"_+Chancellor- of the ~Exchequer-",
	"_+Chancellor- of the ~Duchy of Lancaster-",
	"_+President- of the ~Board of Trade-",
	"_+Minister- Without ~Portfolio-",
	"_*Foreign Secretary-",
	"_*Health Secretary-",
	"_+Minister- for ~Work and Pensions-",
	"_+Minister- for the ~Environment-",
	"_*Deputy Prime Minister-",
	"m+Head- of the Committee for ~Standards in Public Life-",
	"f+Minister- for ~Women-",
	"m*Lord Chancellor-",
	"_@Science and Technology Minister-",
};

#define NUM_MINISTERS ( sizeof(ministers) / sizeof(ministers[0]) )

/* Scandals. Gender-specific words will be substituted for 2-character 
 * sequences that start with a ? mark. */
static char *scandals[] =
{
	"caught with ?h +hand- in the *till-",
	"spending a *lot of nights- with a ?m who is definitely +not- ?h +?w-",
	"stirring up @rumours- of *dissent- in the cabinet",
	"taking +expensive holidays- in *Barbados- instead of working",
	"@lying- to the ~House of Commons-",
	"plagiarising +chunks- of the Leader of the Opposition's *speeches-",
	"receiving +brown envelopes- of grubby *banknotes-",
	"~arrested- for #drunken driving-",
	"seen in a +cheap wine shop-",
	"?!",
};

#define NUM_SCANDALS (sizeof(scandals) / sizeof(scandals[0]))


/* Wbat you can do about the scandal. %s becomes 'him' or 'her'. */
static char *options[] =
{
	"Do #nothing-",
	"+Sack- %s",
	"Issue a ~statement- giving %s your +full support-",
	"Distract the press with a ~controversial policy-",
	"Get %s to apologise tearfully on *national TV-",
	"Order an ~internal enquiry-",
	"Call for a full +public enquiry-",
	"Blame the #Americans-",
	"Blame the #French-",
	"Announce a +tax cut-",
	"Announce *public spending- increases",
	"Deploy +peacekeeping forces- somewhere"
};
#define NUM_OPTIONS (sizeof(options) / sizeof(options[0]))

/* The number of times you've tried a technique. The British public are 
 * impressed by novelty, so the more times you try something, the less likely
 * it is to work. */
static int optcounts[NUM_OPTIONS];


/* Print your final rating */
int result()
{
	char c;

	cls();
	centre(0, "End of game", 1);
	printat(1, 0, "Final popularity: +%d-%%", rating);
/* If rating is below 0, you've been thrown out */
	if (rating <= 0)
	{
		printat(3, 0, "  You have become so #unpopular- that your "
				"own MPs force a vote of +no confidence-. "
				"You *resign- in #disgrace-, and disappear "
				"from public life until you're asked to take "
				"part in the new series of '~I'm a celebrity, "
				"kick me-'.");
	}
	else 
	{
/* Otherwise, you are still in office */
		printat(3, 0, "  You made it to the Summer Recess.");
		if (rating <= 10)
		{
			printat(6, 0, "  People are +marching- with *banners- "
					"calling for you to ~resign-. When "
					"you appear in public #eggs- are "
					"thrown at you.");
		}
		else if (rating < 20)
		{
			printat(6, 0, "  Protesting +farmers- drive their "
				      "~muckspreaders- into central London "
				      "and let fly. The cleanup bill comes "
				      "to *millions-.");
		}
		else if (rating < 30)
		{
			printat(6, 0, "  You are +burnt- in *effigy- by "
				      "demonstrating #sociology students- and "
				      "@golf course managers-.");
		}
		else if (rating < 40)
		{
			printat(6, 0, "  A poll of #young people- reveals "
				      "that most of them have +never heard of "
				      "you-.");
		}
		else if (rating < 50)
		{
			printat(6, 0, "  The ~Women's Institute- greet your "
				      "appearance with +slow handclaps-.");
		}
		else if (rating < 60)
		{
			printat(6, 0, "  A +poll- in the *Daily Telegraph- "
				      "says that people appreciate "
				      "#the smack of firm government-.");
		}
		else if (rating < 70)
		{
			printat(6, 0, "  The *President of Uzbekistan- pays"
					"a fulsome tribute to you on "
					"national television");
		}
		else if (rating < 80)
		{
			printat(6, 0, "  You get invited to be the +guest host-"
					" on a special edition of "
					" #Have I Got News For You-");
		}
/* Getting increasingly unlikely here... */
		else if (rating < 90)
		{
			printat(6, 0, "  England's +cricket team- win the "
					"~ashes-. You reward the captain with "
					"a *peerage-. The nation #celebrates.");
		}
		else
		{
			printat(6, 0, "  You are +knighted- by the *Queen.- "
				      "I also think you *cheated-.");

		}	
	}
	centre(24, "Play again (Y/N)?", 4);
	while(1)
	{
		c = getch();
		if (c == 'y' || c == 'Y') return 1;
		if (c == 'n' || c == 'N') return 0;
	}
	return 0;
}

/* Deal with the scandal of the week. */
void do_week(int week)
{
	char buf[1200];
	char *str, *txt, *append, *deny;
	char g, ch;
	int d, m, s, n, n1;
	int chance;
	int availopts[4];
	int selopt;

/* d is 0 if the minister admits the allegations, nonzero if they deny them */
	d = rand() % 4;

/* Which minister? and which scandal? */
	m = rand() % NUM_MINISTERS;
	s = rand() % NUM_SCANDALS;

	if (d == 0) deny = "admits";
	else        deny = "denies";

	g = toupper(ministers[m][0]);	/* g is the minister's gender */
	str = scandals[s];
	txt = buf;
	/* Generate the scandal text */
	while (*str)
	{
		if (*str == '?')
		{
			++str;
			append = "";
			switch(*str++)
			{
	/* Gender-specific scandal */
				case '!': append = ((g == 'M') ? 
					"'looking for +badgers-' "
					"on *Hampstead Heath-" :
					"consulting a '+lifestyle guru-' who "
					"wears *bangles-");
					  break;
	/* ?h = his / her */
				case 'h': append = ((g == 'M') ? "his" : "her");
					  break;
	/* ?m = woman / man */
				case 'm': append = ((g == 'M') ? "woman" : "man");
					  break;
	/* ?w = wife / husband */
				case 'w': append = ((g == 'M') ? "wife" : "husband");
					  break;
			}
			strcpy(txt, append);
			txt += strlen(append);
		}	
		else *txt++ = *str++;
	}
	*txt = 0;
/* Display current status & scandal */
	cls();
	printat(0, 0, "Week *%d- of your new vision for Britain", week + 1);
	printat(1, 0, "Your approval rating: +%d-%%", rating);

	printat(3, 0, "~Oh no!- The %s has been %s!\n\n"
			"%s %s the allegations.", 
			ministers[m] + 1, buf, (g == 'M') ? "He" : "She",
			deny);

	printat(14, 0, "#What do you do?-");

/* Come up with 4 options from the list */
	memset(availopts, 0, sizeof(availopts));
	for (n = 0; n < 4; )
	{
		availopts[n] = rand() % NUM_OPTIONS;
/* If this option has already been used, try again */
		for (n1 = 0; n1 < n; n1++)
		{
			if (availopts[n1] == availopts[n]) availopts[n] = -1;
		}
		if (availopts[n] >= 0) ++n;
	}
	buf[0] = 0;
/* Display the options */
	for (n = 0; n < 4; n++)
	{
		sprintf(buf + strlen(buf), "%c) ", n + 'a');
		sprintf(buf + strlen(buf), options[availopts[n]], 
				(g == 'M') ? "him" : "her");
		sprintf(buf + strlen(buf), ".\n");
	}
	printat(15, 0, "%s", buf);

/* Get user's choice */
	do
	{
		ch = tolower(getch());
	}
	while (ch < 'a' || ch > 'd');
	selopt = availopts[ch - 'a'];

/* The chance of the option working. It doesn't really matter *what* you do,
 * as long as it's something different each time. And each option is twice
 * as likely to fail as to succeed. */
	chance = (rand() % 12) - (8 + optcounts[selopt]);
	++optcounts[selopt];	

/* If you asked to do nothing, there's a 25% chance of good news and a 25%
 * chance of bad news */	
	if (selopt == 0) 
		switch(rand() % 4)
		{
			case 0:
			case 1: break;
			case 2: 
			printat(21, 0, "A +soap star- decides to get *divorced-.");
			printat(22, 0, "It's all over the newspapers for days.");
			chance += 4;
			break;
			case 3:
			printat(21, 0, "England lose again at football.");
			printat(22, 0, "The newspapers blame your government.");
			chance -= 4;
			break;
		}
/* If you sacked the minister, randomise the gender of the replacement */
	if (selopt == 1) 
	{		
		if (isupper(ministers[m][0]))
		{	
			if (rand() & 3) ministers[n][0] = 'M';
			else		ministers[n][0] = 'F';
		}		
	}

/* Print how you got on */
	if (chance > 0) 
		printat(23, 0, "+Your plan seems to have worked.-");
	else if (chance == 0)
		printat(23, 0, "+Looks like you got away with it.-");
	else	printat(23, 0, "+The public don't seem impressed.-");

	rating += chance;
	centre(24, "Press ENTER to continue", 4);
	getch();
}



int main()
{
	char c;
	int n;
	time_t t;

	/* Take a leaf out of Microsoft's book - display a funky error
	 * message when there's nothing wrong but you aren't running on
	 * an approved system */
	if (!is_sinclair())
	{
		fprintf(stderr, "ES47002: This program is designed for\n"
				"Sinclair PCs. This is not a Sinclair\n"
			        "PC. Y to continue, N to quit.\n");
		do
		{
			c = getch();
			if (c == 'n' || c == 'N') return 1;
		}
		while (c != 'Y' && c != 'y');
	}
	time(&t);
	srand(t);
	initscr();
	intro();
	time(&t);
	srand(t);

	do
	{
		/* Initialise game state */
		rating = 50;
		/* Set up ministers' genders */
		for (n = 0; n < NUM_MINISTERS; n++)
		{
			/* Some ministers have fixed genders */
			if (ministers[n][0] == 'm' || ministers[n][0] == 'f') 
				continue;
			/* Let's give the ministers a 3:1 male:female ratio */
			if (rand() & 3) ministers[n][0] = 'M';
			else		ministers[n][0] = 'F';
		}
		/* Reset the count of how many times you've tried something */
		memset(optcounts, 0, sizeof(optcounts));

		/* Play for 14 weeks */
		for (n = 0; n < 14; n++)
		{
			do_week(n);
			if (rating <= 0) break;
		}
	}
	while (result());
	cls();
	return 0;
}
